// OpenFluidWarmer Gen 2 Prototype
// Safety Companion for Sous Vide Cooker

#include <OneWire.h>
#include <DallasTemperature.h>
#include <TM1637Display.h>

// microcontroller pins
const int one_wire_bus_pin = 2;
const int sous_vide_button_pin = 3;
const int red_LED_pin = 4;
const int set_point_button_pin = 5;
const int blue_LED_pin = 6;
const int display_clk_pin = 7;
const int display_dio_pin = 8;
const int buzzer_pin = 9;
const int relay_pin = 10;
const int iv_tube_switch_pin = 11;

// momentary push buttons
const unsigned long short_button_delay = 100; //milliseconds
const unsigned long long_button_delay = 2000; //milliseconds
int sous_vide_button_prev = HIGH;
unsigned long sous_vide_button_time = 0; //milliseconds
int set_point_button_prev = HIGH;
unsigned long set_point_button_time = 0; //milliseconds
unsigned long set_point_button_delay = short_button_delay; //milliseconds

// iv tube switch
int iv_tube_switch_prev = LOW;

// LEDs
bool red_LED_flash = true;
int red_LED = LOW;
bool blue_LED_flash = true;
int blue_LED = LOW;
const unsigned long LED_flash_on_delay = 10; //milliseconds
unsigned long LED_on_time = 0; //milliseconds
const unsigned long LED_flash_off_delay = 1000; //milliseconds
unsigned long LED_off_time = 0; //milliseconds

// temperature sensors
float temperature_1 = -100.0; //°C
float temperature_2 = -100.0; //°C
float average_temperature = -100.0; //°C
OneWire oneWire(one_wire_bus_pin);
DallasTemperature sensors(&oneWire);
DeviceAddress temp_sense_1, temp_sense_2;
const float temperature_sample_freq = 2.0; //samples per second
unsigned long temperature_sample_time = 0; //milliseconds

// 4 digit, 7 segment display
TM1637Display display(display_clk_pin, display_dio_pin);
const uint8_t symbol_degree = SEG_A | SEG_B | SEG_F | SEG_G;
const uint8_t letter_a = SEG_A | SEG_B | SEG_C |SEG_E | SEG_F | SEG_G;
const uint8_t letter_b = SEG_C | SEG_D |SEG_E | SEG_F | SEG_G;
const uint8_t letter_c = SEG_A | SEG_D | SEG_E | SEG_F;
const uint8_t letter_d = SEG_B | SEG_C |SEG_D | SEG_E | SEG_G;
const uint8_t letter_e = SEG_A | SEG_D |SEG_E | SEG_F | SEG_G;
const uint8_t letter_f = SEG_A | SEG_E | SEG_F | SEG_G; 
const uint8_t letter_h = SEG_C | SEG_E | SEG_F | SEG_G;
const uint8_t letter_l = SEG_D |SEG_E | SEG_F;
const uint8_t letter_n = SEG_C | SEG_E | SEG_G;
const uint8_t letter_o = SEG_A | SEG_B | SEG_C |SEG_D | SEG_E | SEG_F;
const uint8_t letter_p = SEG_A | SEG_B | SEG_E | SEG_F | SEG_G;
const uint8_t letter_s = SEG_A | SEG_C | SEG_D | SEG_F | SEG_G;
const uint8_t letter_t = SEG_D | SEG_E | SEG_F | SEG_G;
const uint8_t letter_u = SEG_B | SEG_C |SEG_D | SEG_E | SEG_F;
const unsigned long display_message_delay = 1000; // milliseconds
unsigned long display_temperature_time = 0; // milliseconds
unsigned long display_fault_time = 0; // milliseconds

// buzzer
int buzzer = LOW;
unsigned long buzzer_time = 0; //milliseconds
const unsigned long buzzer_delay = 50; //milliseconds

// system performance limits
float max_temperature_limit = 75.0; //°C
float max_temperature_difference = 2.0; //°C
float set_point_temperature_tolerance = 0.5; // +/-__°C

// control states
bool sous_vide_power = false;
bool sous_vide_power_prev = false;
bool set_point_monitor = false;
bool set_point_monitor_prev = false;
float set_point_temperature = -100.0; //°C

// faults
int hard_fault = 0;
int soft_fault = 0;

void setup() {
//  Serial.begin(9600);
     
  // initialize pin types
  pinMode(sous_vide_button_pin, INPUT_PULLUP);
  pinMode(red_LED_pin, OUTPUT);
  pinMode(set_point_button_pin, INPUT_PULLUP);
  pinMode(blue_LED_pin, OUTPUT);

  pinMode(buzzer_pin, OUTPUT);
  pinMode(relay_pin, OUTPUT);
  pinMode(iv_tube_switch_pin, INPUT_PULLUP);

  // set initial state of output pins
  digitalWrite(red_LED_pin, LOW);
  digitalWrite(blue_LED_pin, LOW);
  digitalWrite(buzzer_pin, LOW);
  digitalWrite(relay_pin, LOW);

  // temperature sensors
  sensors.begin();
  if (!sensors.getAddress(temp_sense_1, 0)) Serial.println("Unable to find address for Device 0");
  if (!sensors.getAddress(temp_sense_2, 1)) Serial.println("Unable to find address for Device 1");
  sensors.setResolution(temp_sense_1, 9);
  sensors.setResolution(temp_sense_2, 9);

  // 4 digit, 7 segment display
  display.setBrightness(7, true);
}

void loop() {

  // request temperatures
  if (millis() > temperature_sample_time){
    sensors.requestTemperatures();
    temperature_1 = sensors.getTempC(temp_sense_1);
    temperature_2 = sensors.getTempC(temp_sense_2);
//    Serial.print("temperature_1: ");
//    Serial.println(temperature_1);
//    Serial.print("temperature_2: ");
//    Serial.println(temperature_2);
    average_temperature = (temperature_1 + temperature_2)/2.0;
    temperature_sample_time = millis() + (1000/temperature_sample_freq);
  }

  // temperature above max allowable temperature limit fault check
  if (temperature_1 >= max_temperature_limit || temperature_2 >= max_temperature_limit){
    hard_fault = 1;
    sous_vide_power = false;
  }
  else if (hard_fault == 1){
    hard_fault = 0;
  }

  // large temperature difference between temperature sensors fault check
  float temperature_difference = abs(temperature_1 - temperature_2);
  if (temperature_difference >= max_temperature_difference){
    soft_fault = 2;
  }
  else if (soft_fault == 2){
    soft_fault = 0;
  }
  
  // set point monitor fault checks
  if (set_point_monitor){

    // average temperature above set point temperature
    if (average_temperature > set_point_temperature + set_point_temperature_tolerance){
      hard_fault = 2;
      sous_vide_power = false;
    }
    else if (hard_fault == 2){
      hard_fault = 0;
    }

    // average temperature below set point temperature fault check
    if (average_temperature < set_point_temperature - set_point_temperature_tolerance){
      soft_fault = 1;
    }
    else if (soft_fault == 1){
      soft_fault = 0;
    }

    // iv tube switch fault check
    int iv_tube_switch = digitalRead(iv_tube_switch_pin);
    if (iv_tube_switch != iv_tube_switch_prev){
      if (iv_tube_switch == LOW){
        soft_fault = 3;
      }
      else if (iv_tube_switch == HIGH){
        soft_fault = 0;
      }
      iv_tube_switch_prev = iv_tube_switch;
    }
  }

  //display fault when a fault occurs
  if (soft_fault != 0 || hard_fault != 0){
    if (millis() > display_fault_time){
      if (hard_fault == 1){ // above temperature limit
        uint8_t display_data[] = {letter_a, letter_b, letter_t, letter_l};
        display.setSegments(display_data);
      }
      else if (hard_fault == 2){ // average temperature above set point temperature
        uint8_t display_data[] = {letter_a, letter_b, letter_s, letter_p};
        display.setSegments(display_data);
      }
      else if (soft_fault == 1){ // average temperature below set point temperature
        uint8_t display_data[] = {letter_b, letter_l, letter_s, letter_p};
        display.setSegments(display_data);
      }
      else if (soft_fault == 2){ // large temperature sensor difference between temperature sensors
        uint8_t display_data[] = {letter_t, letter_s, letter_d, letter_f};
        display.setSegments(display_data);
      }
      else if (soft_fault == 3){ // iv tube is not present
        uint8_t display_data[] = {letter_t, letter_u, letter_b, letter_e};
        display.setSegments(display_data);
      }
      display_temperature_time = millis() + display_message_delay;
      display_fault_time = millis() + 2*display_message_delay; 

      digitalWrite(buzzer_pin, HIGH);
      buzzer = HIGH;
      buzzer_time = millis() + buzzer_delay;
    }  
    if (millis() > buzzer_time && buzzer == HIGH){
      digitalWrite(buzzer_pin, LOW);
      buzzer == LOW;
    }
  }

  // display temperature continuously
  if (millis() > display_temperature_time){
        
    //temperature display digits
    uint8_t tens = (int(average_temperature + 0.5)/10)%10;
    tens = display.encodeDigit(tens);
    uint8_t ones = int(average_temperature + 0.5)%10;
    ones = display.encodeDigit(ones);
    
    uint8_t display_data[] = {tens, ones, symbol_degree, letter_c};
    display.setSegments(display_data);

    display_temperature_time = temperature_sample_time;
  }
  
  // sous vide button
  int sous_vide_button = digitalRead(sous_vide_button_pin);
  if (sous_vide_button != sous_vide_button_prev){
    if (sous_vide_button == LOW){
      sous_vide_button_time = millis() + short_button_delay;
    }
    else if (sous_vide_button == HIGH && (millis() > sous_vide_button_time)){
      sous_vide_power = !sous_vide_power;
    }
  }
  sous_vide_button_prev = sous_vide_button;

  // set point button
  int set_point_button = digitalRead(set_point_button_pin);
  if (set_point_button != set_point_button_prev){
    if (set_point_button == LOW){
      set_point_button_time = millis() + set_point_button_delay;
    }
    else if (set_point_button == HIGH && (millis() > set_point_button_time)){
      set_point_monitor = !set_point_monitor;
    }
  }
  set_point_button_prev = set_point_button;
  
  // sous vide power
  if (sous_vide_power != sous_vide_power_prev){
    if (sous_vide_power){
      digitalWrite(relay_pin, HIGH);
      red_LED_flash = false;
      digitalWrite(red_LED_pin, HIGH);
      
      uint8_t display_data[] = {letter_h, letter_t, letter_o, letter_n};
      display.setSegments(display_data);
      display_temperature_time = millis() + display_message_delay;
    }
    else{
      digitalWrite(relay_pin, LOW);
      red_LED_flash = true;
      digitalWrite(red_LED_pin, LOW);
      
      uint8_t display_data[] = {letter_h, letter_t, letter_o, letter_f};
      display.setSegments(display_data);
      display_temperature_time = millis() + display_message_delay;
    }
    sous_vide_power_prev = sous_vide_power;
  }

  // set point monitor
  if (set_point_monitor != set_point_monitor_prev){
    if (set_point_monitor){
      set_point_temperature = average_temperature;
      
      blue_LED_flash = false;
      digitalWrite(blue_LED_pin, HIGH);
      set_point_button_delay = long_button_delay;
      
      uint8_t display_data[] = {letter_s, letter_p, letter_o, letter_n};
      display.setSegments(display_data);
      display_temperature_time = millis() + display_message_delay;
    }
    else{
      // get rid of faults related to set point monitoring when set point monitoring is turned off
      if (hard_fault == 2){
        hard_fault = 0;
      }
      if (soft_fault == 1 || soft_fault == 3){
        soft_fault = 0;
        iv_tube_switch_prev = LOW;
      }
      
      blue_LED_flash = true;
      digitalWrite(blue_LED_pin, LOW);
      set_point_button_delay = short_button_delay;
      
      uint8_t display_data[] = {letter_s, letter_p, letter_o, letter_f};
      display.setSegments(display_data);
      display_temperature_time = millis() + display_message_delay;
    }
    set_point_monitor_prev = set_point_monitor;
  }
  
  // red LED flash
  if (red_LED_flash){
    if (red_LED == LOW && (millis() > LED_off_time)){
      digitalWrite(red_LED_pin, HIGH);
      red_LED = HIGH;
      LED_on_time = millis() + LED_flash_on_delay;
    }
    else if(red_LED == HIGH && (millis() > LED_on_time)){
      digitalWrite(red_LED_pin, LOW);
      red_LED = LOW;
      LED_off_time = millis() + LED_flash_off_delay;
    }
  }

  // blue LED flash
  if (blue_LED_flash){
    if (blue_LED == LOW && (millis() > LED_off_time)){
      digitalWrite(blue_LED_pin, HIGH);
      blue_LED = HIGH;
      LED_on_time = millis() + LED_flash_on_delay;
    }
    else if(blue_LED == HIGH && (millis() > LED_on_time)){
      digitalWrite(blue_LED_pin, LOW);
      blue_LED = LOW;
      LED_off_time = millis() + LED_flash_off_delay;
    }
  }

}
